-- MT46V16M16TG-5B
library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity sdram is
		generic (sim : boolean := false);
    port (
-- reset input
        reset: in STD_LOGIC;
        rst_DCM: in STD_LOGIC;
-- system clock input
        clk: in STD_LOGIC; -- 100MHz 
        clk_out: out STD_LOGIC; 
        clk2x_out: out STD_LOGIC; 
        mem_status : out  STD_LOGIC_VECTOR (15 downto 0);
-- input interface signals
-- ddr write queue signals
				wr_rqst: in std_logic;
				wr_gnt: out std_logic;
        wr_addr: in STD_LOGIC_VECTOR (22 downto 0); -- wr_addr must be all zero when mem_status(14) = '0'
        wr_length: in STD_LOGIC_VECTOR (5 downto 0);
        wr_data: in STD_LOGIC_VECTOR (31 downto 0);
				wr_en: in std_logic;
-- output interface signals
        rd_rqst: in STD_LOGIC;
        rd_gnt: out STD_LOGIC;
        rd_addr: in STD_LOGIC_VECTOR (22 downto 0);
        rd_length: in STD_LOGIC_VECTOR (5 downto 0);
        rd_en: in STD_LOGIC;
        rd_data: out STD_LOGIC_VECTOR (31 downto 0);
-- ram pads
				mdq : inout STD_LOGIC_VECTOR (15 downto 0);
				maddr : out STD_LOGIC_VECTOR (12 downto 0);
				mba : out STD_LOGIC_VECTOR (1 downto 0);
				dm : out STD_LOGIC_VECTOR (1 downto 0);
				dqs : inout STD_LOGIC_VECTOR (1 downto 0);
				we_n : out STD_LOGIC;
				ras_n : out STD_LOGIC;
				cas_n : out STD_LOGIC;
				cs_n : out STD_LOGIC;
				mclke : out STD_LOGIC;
				mclk : out STD_LOGIC;
				mclk_n : out STD_LOGIC
    );
end sdram;

architecture sdram_arch of sdram is
signal clk2x: std_logic :='0';
signal clkn: std_logic :='0';
-- sdram i/o
signal ma: std_logic_vector(14 downto 0) := (others => '0');
signal md: std_logic_vector(31 downto 0) := (others => '0');
signal mq: std_logic_vector(31 downto 0) := (others => '0');
signal mq_q: std_logic_vector(31 downto 0) := (others => '0');
signal active: std_logic :='0';
signal rasn: std_logic :='0';
signal casn: std_logic :='0';
signal wen: std_logic :='0';
signal active_dl: std_logic_vector(3 downto 0) := (others => '0');
signal t_md: std_logic :='0';
signal t_dqs: std_logic :='0';
signal d_dm: std_logic :='0';
-- refresh signal generation
signal ref: std_logic :='0';
signal start_init: std_logic :='0';
-- sdram state machine
signal is_read: std_logic :='0';
signal rqst: std_logic :='0';
signal EMR: std_logic := '0';
signal MR: std_logic := '0';
signal w_cycle: std_logic :='0';
signal r_cycle: std_logic :='0';
signal burst_cntr: std_logic_vector(6 downto 0) := (others => '0');
signal rw_burst: std_logic :='0';
signal wr_burst: std_logic :='0';
signal rd_burst: std_logic :='0';
signal rd_burstp: std_logic :='0';
signal rd_burst_dl: std_logic :='0';
signal stop_w: std_logic :='0';
signal stop_dqs: std_logic :='0';
signal stop_r: std_logic :='0';
signal inh_ras: std_logic :='0';
signal ref_rqst: std_logic :='0';
signal init_done: std_logic_vector(1 downto 0) := (others => '0');
signal pc_dl: std_logic :='0';
signal precharge: std_logic :='0';
signal precharge_dl: std_logic :='0';
signal rfc_cntr: std_logic_vector(3 downto 0) := x"0";
signal mclk0_dcm: std_logic :='0';
signal mclk90_dcm: std_logic :='0';
--signal mclk180_dcm: std_logic :='0';
--signal mclk270_dcm: std_logic :='0';
signal mclk0: std_logic :='0';
signal mclk90: std_logic :='0';
signal mclk180: std_logic :='0';
signal mclk270: std_logic :='0';
signal lock: std_logic :='0';
signal ec_wfifo_ra: std_logic :='0';
signal wfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal wfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal wfifo_dout : std_logic_vector(31 downto 0) := (others => '0');
signal rfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal dqs_in : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_in_q : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_status : std_logic_vector(3 downto 0) := (others => '0');
signal rfifo_we: std_logic := '0';
signal psincdec: std_logic := '0';
signal ps_delay : std_logic_vector(7 downto 0) := (others => '0');
signal psen: std_logic := '0';
signal psdone: std_logic := '0';
signal ps_rdy: std_logic := '0';
signal ps_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal ps_tgl: std_logic := '0';
signal ps_tgl_sync : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_out: std_logic :='0';
signal mem_rdy: std_logic :='0';
signal mem_rdyp: std_logic :='0';
begin
clk_out <= mclk0;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    if(psdone = '1')then
      mem_status(7 downto 0) <= ps_cntr;
		end if;
		mem_status(15 downto 8) <= mem_rdy & init_done(1) & ps_rdy & "00000";
  end if;
end process;  
clkn <= not clk;
i_mclk:  fddrrse
    port map(
          d0 => '1',
          d1 => '0',
          c0 => clkn,
          c1 => clk,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mclk);
i_mclk_n:  fddrrse
    port map(
          d0 => '0',
          d1 => '1',
          c0 => clkn,
          c1 => clk,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mclk_n);
DCMmclk: dcm 
   generic map(CLKOUT_PHASE_SHIFT => "VARIABLE",
					CLKIN_PERIOD => 10.0,
					PHASE_SHIFT => 0)  
	port map	
         ( CLKIN   => clk ,
           CLKFB   => mclk90 ,
           RST     => rst_DCM ,
           DSSEN   => '0' ,
           PSINCDEC=> psincdec ,
           PSEN    => psen ,
           PSCLK   => mclk0 ,
           PSDONE  => psdone ,
           CLK2X    => clk2x ,
           CLK0    => mclk90_dcm ,
           CLK270  => mclk0_dcm ,
           LOCKED  => lock);
clk2xbufg : BUFG port map (O=>clk2x_out, I=>clk2x);
mclk0bufg : BUFG port map (O=>mclk0, I=>mclk0_dcm);
mclk90bufg : BUFG port map (O=>mclk90, I=>mclk90_dcm);
mclk180 <= not mclk0;
mclk270 <= not mclk90;
process(mclk0,lock)
begin
  if(lock = '0')then
    ps_rdy <= '1';
    ps_cntr <= x"80";
	  dqs_status <= x"0";
	  psen <= '0';
	  psincdec <= '0';
		if(sim = true)then
			ps_delay <= x"7f";
		else
			ps_delay <= x"00";
		end if;
		mem_rdyp <= '0';
  elsif(mclk0'event and mclk0 = '1')then
    ps_rdy <= not psen and (ps_rdy or psdone);
		if(psen = '1')then
			if(psincdec = '1')then
				ps_cntr <= ps_cntr + 1;
			else
				ps_cntr <= ps_cntr - 1;
			end if;
		end if;
		if(psdone = '1' or ref = '1')then
			dqs_status <= x"0";
		elsif(rfifo_we = '1')then
			dqs_status(0) <= dqs_status(0) or not dqs_in_q(0);
			dqs_status(1) <= dqs_status(1) or dqs_in_q(0);
			dqs_status(2) <= dqs_status(2) or not dqs_in_q(1);
			dqs_status(3) <= dqs_status(3) or dqs_in_q(1);
		end if;
		if(ref = '1' and ps_rdy = '1')then
			case dqs_status is
				when "1001" =>
					psen <= '1';
					psincdec <= '1';
				when "0110" =>
					psen <= '1';
					psincdec <= '0';
				when others => null;	 
			end case;
		else
			psen <= '0';
		end if;
		if(ps_delay(7) = '0' and stop_r = '1')then
			ps_delay <= ps_delay + 1;
		end if;
		if(ref = '1')then
		  mem_rdyp <= mem_rdy;
		end if;
  end if;
end process;
mem_rdy <= ps_delay(7);
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    dqs_in_q <= dqs_in;
  end if;
end process;
-- ram pads
i_dqs0: ddr_io
    port map(
          clk => mclk270,
          clkn => mclk90,
          pad => dqs(0),
					t => t_dqs,
          d0 => dqs_out,
          d1 => '0',
          iclk => mclk0,
          iclkn => mclk180,
          q0 => dqs_in(0),
          q1 => dqs_in(1));
i_dqs1: ddr_io
    port map(
          clk => mclk270,
          clkn => mclk90,
          pad => dqs(1),
					t => t_dqs,
          d0 => dqs_out,
          d1 => '0',
          iclk => '0',
          iclkn => '0');
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    we_n <= wen;
    cas_n <= casn;
    ras_n <= rasn;
    maddr <= ma(12 downto 0);
		mba <= ma(14 downto 13);
  end if;
end process;
g_mdq: for i in 0 to 15 generate
  ddr_data: ddr_io
    port map(
          clk => mclk0,
          clkn => mclk180,
          pad => mdq(i),
					t => t_md,
          d0 => md(16+i),
          d1 => md(i),
          iclk => mclk90,
          iclkn => mclk270,
          q0 => mq(i),
          q1 => mq(16+i));
end generate g_mdq;
process(mclk90)
begin
  if(mclk90'event and mclk90 = '1')then
    mq_q <= mq;
  end if;
end process;
md(15 downto 0) <= wfifo_dout(15 downto 0);
process(mclk180)
begin
  if(mclk180'event and mclk180 = '1')then
    md(31 downto 16) <= wfifo_dout(31 downto 16);
  end if;
end process;
i_wfifo : RAMB16BWE_S36_S36
   port map (
      DOB => wfifo_dout,      -- Port B 32-bit Data Output
      ADDRA => wfifo_wa,  -- Port A 9-bit Address Input
      ADDRB => wfifo_ra,  -- Port B 9-bit Address Input
      CLKA => mclk0,    -- Port A 1-bit Clock
      CLKB => mclk180,    -- Port B 1-bit Clock
      DIA => wr_data,      -- Port A 32-bit Data Input
      DIB => x"00000000",      -- Port B 32-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
      ENA => wr_en,      -- Port A 1-bit RAM Enable Input
      ENB => '1',      -- Port B 1-bit RAM Enable Input
      SSRA => '0',    -- Port A 1-bit Synchronous Set/Reset Input
      SSRB => '0',    -- Port B 1-bit Synchronous Set/Reset Input
      WEA => x"f",      -- Port A 4-bit Write Enable Input
      WEB => x"0"       -- Port B 4-bit Write Enable Input
   );
process(mclk180,reset)
begin
  if(reset = '1')then
	  wfifo_ra <= (others => '0');
		ec_wfifo_ra <= '0';
  elsif(mclk180'event and mclk180 = '1')then
	  if(ec_wfifo_ra = '1')then
		  wfifo_ra <= wfifo_ra + 1;
		end if;
		ec_wfifo_ra <= wr_burst;
  end if;
end process;  
i_rfifo : RAMB16BWE_S36_S36
   port map (
      DOB => rd_data,      -- Port B 32-bit Data Output
      ADDRA => rfifo_wa,  -- Port A 9-bit Address Input
      ADDRB => rfifo_ra,  -- Port B 9-bit Address Input
      CLKA => mclk90,    -- Port A 1-bit Clock
      CLKB => mclk0,    -- Port B 1-bit Clock
      DIA => mq_q,      -- Port A 32-bit Data Input
      DIB => x"00000000",      -- Port B 32-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
      ENA => rfifo_we,      -- Port A 1-bit RAM Enable Input
      ENB => '1',      -- Port B 1-bit RAM Enable Input
      SSRA => '0',    -- Port A 1-bit Synchronous Set/Reset Input
      SSRB => '0',    -- Port B 1-bit Synchronous Set/Reset Input
      WEA => x"f",      -- Port A 4-bit Write Enable Input
      WEB => x"0"       -- Port B 4-bit Write Enable Input
   );
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(reset = '1')then
			wfifo_wa <= (others => '0');
    elsif(wr_en = '1')then
			wfifo_wa <= wfifo_wa + 1;
		end if;
  end if;
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    if(reset = '1')then
      rfifo_ra <= (others => '0');
		elsif(rd_en = '1')then
			rfifo_ra <= rfifo_ra + 1;
		end if;
  end if;
end process;
i_rd_burst_dl: delayN
  generic map(N=> 4) port map(d => rd_burstp, ce => '1', c => mclk90, q => rd_burst_dl);
process(mclk90,reset)
begin
  if(reset = '1')then
		rfifo_we <= '0';
    rfifo_wa <= (others => '0');
  elsif(mclk90'event and mclk90 = '1')then
		rfifo_we <= rd_burst_dl;
    if(rfifo_we = '1' and mem_rdyp = '1')then
			rfifo_wa <= rfifo_wa + 1;
		end if;
  end if;
end process;
process(mclk180,reset)
begin
  if(reset = '1')then
		rd_burstp <= '0';
		dm <= "00";
  elsif(mclk180'event and mclk180 = '1')then
		rd_burstp <= rd_burst;
		dm <= d_dm & d_dm;
  end if;
end process;
-- ram access request logic
process(mclk0,reset)
begin
  if(reset = '1')then
    rqst <= '0';
		wr_gnt <= '0';
		rd_gnt <= '0';
  elsif(mclk0'event and mclk0 = '1')then
    if(precharge = '1' or stop_r = '1')then
      rqst <= '0';
    elsif(wr_rqst = '1' or rd_rqst = '1' or (init_done(1) = '1' and mem_rdy = '0' and ref_rqst = '1'))then
      rqst <= '1';
    end if;
		wr_gnt <= active_dl(0) and w_cycle;
		rd_gnt <= active_dl(0) and r_cycle and mem_rdy;
  end if; 
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    if(rqst = '0')then
			is_read <= not wr_rqst and init_done(1);
    end if;
-- this ensures active, read/write and precharge are to the same bank
    if(inh_ras = '0')then
      if(is_read = '0')then
        ma(14) <= wr_addr(22);
        ma(13) <= wr_addr(21) or EMR;
			else
				ma(14 downto 13) <= rd_addr(22 downto 21);
			end if;
		end if;
		if(inh_ras = '0')then
      if(is_read = '0')then
        ma(12 downto 11) <= wr_addr(20 downto 19);
        ma(10) <= wr_addr(18) or (precharge and not init_done(1));
        ma(9) <= wr_addr(17);
-- reset dll
        ma(8) <= wr_addr(16) or MR;
        ma(7 downto 6) <= wr_addr(15 downto 14);
-- latency = 2
        ma(5) <= wr_addr(13) or MR;
        ma(4 downto 2) <= wr_addr(12 downto 10);
-- burst = 8
        ma(1) <= wr_addr(9) or MR;
        ma(0) <= wr_addr(8) or MR;
			else  
				ma(12 downto 0) <= rd_addr(20 downto 8);
			end if;
		elsif(active_dl(0) = '1')then
      if(is_read = '0')then
        ma(12 downto 0) <= "00" & precharge & '0' & wr_addr(7 downto 0) & '0';
			else  
        ma(12 downto 0) <= "00" & stop_r & '0' & rd_addr(7 downto 0) & '0';
			end if;
		elsif(rw_burst = '1')then
			ma(6 downto 1) <= ma(6 downto 1) + 1;
		end if;
		rd_burst <= burst_cntr(6) and r_cycle;
		wr_burst <= burst_cntr(6) and w_cycle;
		rw_burst <= burst_cntr(6);
		active_dl <= active_dl(2 downto 0) & active;
  end if;
end process;
-- generate read and write burst signals -------------------------------------------------------
process(mclk0,reset)
begin
  if(reset = '1')then
    t_md <= '1';
		d_dm <= '0';
		burst_cntr <= (others => '0');
  elsif(mclk0'event and mclk0 = '1')then
    t_md <= not(not casn and w_cycle) and (t_md or pc_dl);
		d_dm <= pc_dl or (d_dm and casn);
--    if(rqst = '1' and ref_rqst = '0' and inh_ras = '0' and rasn = '1')then
    if(active = '1')then
		  if(is_read = '0')then
			  burst_cntr <= '1' & not wr_length;
			else
			  burst_cntr <= '1' & not rd_length;
			end if;
		elsif(burst_cntr(6) = '1')then
		  burst_cntr <= burst_cntr + 1;
		end if;
  end if;
end process;
process(mclk270,reset)
begin
  if(reset = '1')then
    t_dqs <= '1';
		dqs_out <= '0';
  elsif(mclk270'event and mclk270 = '1')then
    t_dqs <= not(not casn and w_cycle) and (t_dqs or stop_dqs);
		dqs_out <= not t_dqs;
	end if;
end process;
i_stop_dqs: delayN
  generic map(N=> 4) port map(d => stop_w, ce => '1', c => mclk0, q => stop_dqs);
-- sdram state machine -----------------------------------------------------------------------------------------
-- sdram power on initialization states: precharge all banks, enable DLL, reset DLL, precharge all banks,
-- two refresh cycle and set mode register.
i_EMR: delayN
  generic map(N=> 4) port map(d => start_init, ce => '1', c => mclk0, q => EMR);
i_MR: delayN
  generic map(N=> 6) port map(d => start_init, ce => '1', c => mclk0, q => MR);
process(mclk0,lock)
variable state: std_logic_vector(2 downto 0);
begin
state := rasn & casn & wen;
  if(lock = '0')then
		stop_w <= '0';
		stop_r <= '0';
    w_cycle <= '0';
    r_cycle <= '0';
    pc_dl <= '0';
    precharge <= '0';
    precharge_dl <= '0';
    rfc_cntr <= "0000";
		inh_ras <= '0';
		active <= '0';
    rasn <= '1';
    casn <= '1';
    wen <= '1';
    init_done <= "00";
    ref_rqst <= '0';
  elsif(mclk0'event and mclk0 = '1')then
--		if(burst_cntr(6) = '1' and burst_cntr6_q = '0')then
--			stop_w <= w_cycle;
--		else
--			stop_w <= '0';
--		end if;
		stop_w <= wr_burst and not burst_cntr(6);
--		if(burst_cntr(6) = '1' and burst_cntr6_q = '0')then
		if(stop_r = '0' and rd_burst = '1' and (and_reduce(burst_cntr) = '1' or burst_cntr(6) = '0'))then
			stop_r <= '1';
		else
			stop_r <= '0';
		end if;
		pc_dl <= start_init or MR or stop_w;
    precharge <= pc_dl or stop_r;
    precharge_dl <= precharge;
-- auto refresh tRFC counter
    if(state = "001")then
			rfc_cntr(3) <= '1';
		elsif(rfc_cntr(2) = '1')then
			rfc_cntr(3) <= '0';
		end if;
		if(rfc_cntr(3) = '0')then
			rfc_cntr(2 downto 0) <= "000";
		else
			rfc_cntr(2 downto 0) <= rfc_cntr(2 downto 0) + 1;
		end if;
-- inhibit RAS after refresh or read/write until tRFC or tPR
    inh_ras <= not precharge_dl and not (rfc_cntr(2) and rfc_cntr(0)) and ((not rasn and wen) or inh_ras) ;
-- latch refresh request
    if(ref = '1')then
      ref_rqst <= '1';
    end if;
    if(rqst = '1' and ref_rqst = '0' and inh_ras = '0' and rasn = '1')then
		  active <= '1';
		else
		  active <= '0';
		end if;
--		active <= not rasn and casn and wen;
--    if(state = "011")then
--			w_cycle <= not is_read;
--		elsif(stop_w = '1')then
--			w_cycle <= '0';
--		end if;
    if(active = '1')then
			w_cycle <= not is_read;
			r_cycle <= is_read;
		elsif(rasn = '0')then
			r_cycle <= '0';
			w_cycle <= '0';
		end if;
    case state is
-- NOP
      when "111" =>
---- goto precharge
        if(precharge = '1')then
          rasn <= '0';
          wen <= '0';
        end if;
---- goto auto refresh
        if(ref_rqst = '1' and inh_ras = '0')then
          rasn <= '0';
          casn <= '0';
        end if;
---- goto set mode
        if(EMR = '1' or MR = '1')then
          rasn <= '0';
          casn <= '0';
          wen <= '0';
        end if;
---- goto active
        if(rqst = '1' and inh_ras = '0')then
          rasn <= '0';
        end if;
---- goto write/read
        if(active_dl(0) = '1' or (burst_cntr(6) = '1' and ma(2 downto 1) = "11"))then
          casn <= '0';
          wen <= is_read;
        end if;
-- AUTO REFRESH
      when "001" =>
				ref_rqst <= '0';
        init_done <= init_done(0) & '1';
				rasn <= '1';
				casn <= '1';
      when others =>
				rasn <= '1';
				casn <= '1';
				wen <= '1';
    end case;
  end if;
end process;
----------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------
-- instantiate refresh generator
I_refresh: refresh PORT MAP(
		clk => mclk0,
		lock => lock,
		init_done => init_done(1),
		start_init => start_init,
		ddr_csn => cs_n,
		ddr_clke => mclke,
		ref => ref
	);
end sdram_arch;
